# ******************************************************************************************************
# Written by Roberto Ricca in 2023
#
# Modified by Alexander Lenhard in 2024, to make Cuts available in "Schachturnierorganisationsprogramm"!
# ******************************************************************************************************

# Input file is a TRF16 file 
# Usually called with: gawk -f allTB.awk [-v show=VAL] DIR\TRNTRF.trf

function ptsort(lst, elems, aux) {
	for (k=1; k<=elems; k++) {
		max = k
		for (n=k+1; n <= elems; n++) {
			if (lst[max] >= lst[n]) continue
			max = n
		}
		# switch k and max
		sav = lst[k]
		lst[k] = lst[max]
		lst[max] = sav
		if (aux[1] != "") {
			sav = aux[k]
			aux[k] = aux[max]
			aux[max] = sav
		}
		
	}
}

BEGIN {
}


"" $1 != "001" {next}

# Recognized a Player record

# For each player, it is collected
# = the pairing number (pn)
# = the score (score - just to check)
# = (for each round) the opponent (if any) and the result
# = the number of games played
{
	pn = 0 + substr($0, 5, 4)
	trfline[pn] = $0
	score[pn] = 0 + substr($0, 80, 5)
	rnds = int((length - 89) / 10)
	if (rnds > totRounds) totRounds = rnds
	points = 0
	zhf = 0
	line[pn] = ""
	for (r = 1; r <= rnds; r++) {
		opn = 0 + substr($0, 82 + r * 10, 4)
		res = "" substr($0, 89 + r * 10, 1)
		line[pn] = line[pn] " # " sprintf("%2d%2s", opn, res) "  "
		if (res ~ /[1WwFf]/) {
			val = 1  # win
		} else if (res == "+") {
			val = 1  # forfeit win - no opp
			opn = 0
		} else if (res ~ /[=Dd]/) {
			val = 0.5  # draw
		} else if (res ~ /[Hh]/) {
			val = 0.5  # half-point-bye
			opn = -1 # VUR
		} else if (res ~ /[0Ll]/) {
			val = 0  # loss
		} else if (res ~ /[ Zz]/) {
			val = 0  # zero-point-bye
			opn = -2 # VUR
		} else if (res == "-") {
			val = 0  # forfeit loss
			opn = -2 # VUR
		} else if (res ~ /[Uu]/) {
			if (pab == 0) val = 1; else val = pab
		} else {
			print "WRONG RESULT AT LINE " NR "; char=" 90+r*10
			exit 1
		}
		if (opn >= 0) zhf = 0; else if (res ~ /[ Zz]/) zhf++
		result[pn "@" r] = val
		#print r, res, val
		opponent[pn "@" r] = opn
		points += val
	}
	if (points != score[pn]) {
		print "WRONG SCORE FOR " pn " - READ:" score[pn] "; COMP:" points
		exit 1
	}
	adjust[pn] = points + 0.5 * zhf
	#printf "ID=%3d PTS=%3.1f ADJ=%3.1f\n", pn, points, adjust[pn]
	
	#if (pn == 1) for (x=1; x<=totRounds; x++) {
	#	printf "%2d OPP=%2d RES=%3.1f\n", pn, opponent[pn "@" x], result[pn "@" x]
	#}
}

END {
	players = pn
	# any number is ok - computes BH BH-C1 BH-C2
	for (pn=1; pn<=players; pn++) {
		# if (show == 0 || and(show,4)) print trfline[pn]
		# Need two lists STD and VUR
		# For SB, determine the least significant value (absolute) and the lowest
		# VUR contribute, and cut the highest of the two (repeat for SB Cut-2)
		# So we need: a list of bases (SBAUX, equal to STD+VUR); a list of scores
		# (SBSCR to be paired to SBAUX), and a list of VUR contributes (SBVUR)
		std = 0
		vur = 0
		sbx = 0
		sbvur = 0
		sbTot = 0 # just for checking
		# if (show == 0 || and(show,2)) printf "%2d:", pn
		for (r=1; r<=totRounds; r++) {
			opp = opponent[pn "@" r]
			if (opp == "") {
				# undefined so it is ZPN (VUR -2)
				opp = -2
			}
			mark = " " 
			if (opp == 0) {
				sbval = STD[++std] = score[pn]
			} else if (opp < 0) {
				sbval = VUR[++vur] = score[pn]
			} else {
				sbval = STD[++std] = adjust[opp]
				if (adjust[opp] != score[opp]) mark = "*"
			}
			# SB-management
			SBAUX[++sbx] = sbval
			valres = result[pn "@" r]
			tmpx = sbval * valres
			if (valres == 1) sym = "w"; else if (valres == 0) sym = "l"; else sym = "d"
			SBSCR[sbx] = tmpx
			if (opp < 0) SBVUR[++sbvur] = tmpx
			sbTot += tmpx
			# if (show == 0 || and(show,2)) printf "%4d %1s%5.2f%1s ", opp, sym, sbval, mark 
		}
		# if (show == 0 || and(show,2)) print ""
		# checks
		if (std + vur != totRounds || sbx != totRounds) {
			print "FAILURE: std+vur=" std+vur "; sbx=" sbx "; totRounds=" totRounds
		}
		ptsort(STD, std)
		ptsort(VUR, vur)
		ptsort(SBAUX, sbx, SBSCR)
		ptsort(SBVUR, sbvur)
		bh = 0
		sb = 0
		med = 0
		for (b=1; b<=totRounds; b++) {
			sv = bh
			if (b <= std)
				temp = STD[b]
			else
				temp = VUR[b-std]
			bh += temp
			sb += SBSCR[b]
			# print b, bh - sv, bh
			# CUTS
			# print b, totRounds 
			if (b == totRounds - 2) {
				bhC2 = bh
			} else if (b == totRounds - 1) {
				bhC1 = bh
				promed = temp
			}
			if (b <= totRounds - 2) {
				MED[++med] = temp
			}
		}
		# Sonneborn-Berger
		if (sb != sbTot) {
			print "ERROR IN SB: correct=" sbTot "; wrong=" sb
			exit 1
		}
		sblsv = SBSCR[sbx]
		if (sbvur > 0) sblowVUR = SBVUR[sbvur]; else sblowVUR = -1
		if (sblsv < sblowVUR) {
			sbCut1 = sbTot - sblowVUR
			# sblsv is still the lowest significant value, so:
			if (sbvur > 1) sblowVUR = SBVUR[sbvur - 1]; else sblowVUR = -1 
		} else {
			sbCut1 = sbTot - sblsv
			# sblowVUR is still the lowest VUR contribute
			sblsv = SBSCR[sbx - 1]
		}
		if (sblsv < sblowVUR) {
			sbCut2 = sbCut1 - sblowVUR 
		} else {
			sbCut2 = sbCut1 - sblsv
		}

		# Median
		# All values have been collected in MED (cut last two values) 
		# and promed1 (to be added to median to get array wth cut-1)
		ptsort(MED, med)
		median2 = bhC2 - MED[1] - MED[2]
		MED[++med] = promed
		ptsort(MED, med + 1)
		median1 = bhC1 - MED[1]
		# print line[pn]
		# if (show == 0 || and(show,1)) {
		#	printf "%4d %4.1f %4.1f (STD=%2d VUR=%2d)    BH=%5.1f   BH-1=%5.1f   BH-2=%5.1f   BHm1=%5.1f   BHm2=%5.1f   SB=%5.2f   SB-1=%5.2f   SB-2=%5.2f\n", pn, score[pn], adjust[pn], std, vur, bh, bhC1, bhC2, median1, median2, sbTot, sbCut1, sbCut2
		# }

		if (show == 1 && and(show,1)) {
			printf "%5.1f\n", bhC1
		}

		if (show == 2 && and(show,2)) {
			printf "%5.1f\n", bhC2
		}

		if (show == 3 && and(show,3)) {
			printf "%5.1f\n", median1
		}

		if (show == 4 && and(show,4)) {
			printf "%5.1f\n", median2
		}

		if (show == 5 && and(show,5)) {
			printf "%5.2f\n", sbCut1
		}

		if (show == 6 && and(show,6)) {
			printf "%5.2f\n", sbCut2
		}

		# print ""
	}
}